# Control software Raspberry Pico servo
# Details on:
# https://homofaciens.de/technics-physical-computing-PICO-servo_en.htm

from machine import Pin, PWM, I2C
from ssd1306 import SSD1306_I2C
import framebuf,sys
import time

from rp2 import PIO, StateMachine, asm_pio
from machine import Pin
import utime

# ++++++++++ Pin definitions +++++++++++++
HALL_1_PIN = 16
HALL_2_PIN = 17
HALL_3_PIN = 12

PWM_READ = 11

MOTOR_EN_PIN = 13
MOTOR_PWM_PIN = 14
MOTOR_DIR_PIN = 15

OLED_SCL_PIN = 19
OLED_SDA_PIN = 18

LED_INT_PIN = 25 # Internal LED on Pico


# +++++++++ Parameters ++++++++++++++
MAX_DUTY_16 = 65025 # # Max value of duty_u16

PWM_MAX = 50000 # Maximum PWM for motor (MAX_DUTY_16 or lower)
PWM_MIN = 7000  # Minimal PWM needed to move the motor
PWM_END_SWITCH = 10000 # PWM while moving to end switch

SET_POINT_180 = 700 # Number of pulses for 180 degree rotationl

SET_POINT_MAX = 700 # Max position of the servo horn

CONSTANT_ROTATION = False # If true, servo is in conituous rotation mode
READ_SERVO_PWM = True    # If true, PWM control signal is used (default)

END_SWITCH_LED = True    # if true, internal LED is end switch indicator

# ignores servo control signals and instead goes to different
# servo positions. Set True only for testing the servo !
TEST_SERVO_POS = False

# display states of sensors on motor instead of other parameters
DISPLAY_SENSOR_STATE = False;

# The higher the proportional factor, the lower the servo speed
# but the less overshooting appears
PID_PROPORTIONAL = 150 # default: 100

# Plus/Minus Range of accepted setpoint
# the higher, the less overshooting
# the lower the more precisely the set point is reached
SET_POINT_TOLERANCE = 2 # default: 2

# ++++++++++++ Init variables ++++++++++++++++
encoderSteps = 0
encoderStepsOld = 0

pwm_set = 0

# Speed in constant rotation mode
speed_set = 24  # default: 24
current_speed = 0
speed_old = 0
old_speed_diff = 0   # Difference last speed measurement
current_speed_diff = 0   # Current difference in speed measurement

oled_x = 0     # start point for text in x-dir
oled_y = 0     # start point for text in y-dir
oled_line = 10 # space between text in y-dir

hall_1_old = 0

# Angle servo horn in servo mode
set_point = 0

# Pause counter between two servo positions
# if TEST_SERVO_POS is True
new_servo_pos = 0

# value for pulse length from servo control pin
servo_ms = 0

# Counter for bad servo condition
OutOfRange = 0

# ++++++++++++ OLED ++++++++++++++++++++
pix_res_x  = 128 # SSD1306 OLED horizontal resolution
pix_res_y = 64   # SSD1306 OLED vertical resolution

# Init OLED pins
i2c_dev = I2C(1,scl=Pin(OLED_SCL_PIN),sda=Pin(OLED_SDA_PIN),freq=400000)
# get I2C address in hex format
i2c_addr = [hex(ii) for ii in i2c_dev.scan()]
if i2c_addr==[]:
    print('No I2C Display Found')
    sys.exit() # exit routine if no dev found
else:
    print("I2C Address      : {}".format(i2c_addr[0])) # I2C device address
    print("I2C Configuration: {}".format(i2c_dev))     # print I2C params


# ++++++++++ create objects ++++++++++++++
oled = SSD1306_I2C(pix_res_x, pix_res_y, i2c_dev)      # oled controller

led_int = Pin(LED_INT_PIN, Pin.OUT)

motor_EN = Pin(MOTOR_EN_PIN, Pin.OUT)
motor_PWM = PWM(Pin(MOTOR_PWM_PIN))
motor_PWM.freq(1000)
motor_DIR = Pin(MOTOR_DIR_PIN, Pin.OUT)

motor_DIR.low()
motor_PWM.duty_u16(0)

hall_1 = Pin(HALL_1_PIN, Pin.IN, Pin.PULL_UP)
hall_2 = Pin(HALL_2_PIN, Pin.IN, Pin.PULL_UP)
hall_3 = Pin(HALL_3_PIN, Pin.IN, Pin.PULL_UP)

PWM_read = Pin(PWM_READ, Pin.IN, Pin.PULL_UP)
servo_pulse = 0

hall_1_state = hall_1.value()
hall_2_state = hall_2.value()

hall_state = hall_1.value() + 10 * hall_2.value()

# ++++++++ Definitions of sub routines ++++++++++

def to_int32(val): # Transform encoderSteps to signed integer
    val &= ((1<<32)-1)
    if val & (1<<31): val -= (1<<32)
    return val


def sensors_01(): # Check sensors and display states on OLED
    if hall_1.value() and hall_2.value():
          #if not (hall_1_old == 1 and hall_2_old == 1):
          oled.fill(0)
          oled.text("   XX     XX  ",0,0)
          oled.text("  XXX    XXX  ",0,8)
          oled.text(" X  X   X  X  ",0,16)
          oled.text("    X      X  ",0,25)
          oled.text("    X      X  ",0,32)
          oled.text("    X      X  ",0,40)
          oled.text("    X      X  ",0,48)
          oled.text("  XXXXX  XXXXX",0,56)
          oled.show()
          hall_1_old = 1
          hall_2_old = 1

    if not hall_1.value() and not hall_2.value():
          #if not (hall_1_old == 0 and hall_2_old == 0):
          oled.fill(0)
          oled.text("   XX     XX  ",0,0)
          oled.text("  X  X   X  X ",0,8)
          oled.text(" X    X X    X",0,16)
          oled.text(" X    X X    X",0,24)
          oled.text(" X    X X    X",0,32)
          oled.text(" X    X X    X",0,40)
          oled.text("  X  X   X  X ",0,48)
          oled.text("   XX     XX  ",0,56)
          oled.show()
          hall_1_old = 0
          hall_2_old = 0

    if hall_1.value() and not hall_2.value():
          #if not (hall_1_old == 1 and hall_2_old == 0):
          oled.fill(0)
          oled.text("   XX     XX    ",0,0)
          oled.text("  XXX    X  X   ",0,8)
          oled.text(" X  X   X    X  ",0,16)
          oled.text("    X   X    X  ",0,25)
          oled.text("    X   X    X  ",0,32)
          oled.text("    X   X    X  ",0,40)
          oled.text("    X    X  X   ",0,48)
          oled.text("  XXXXX   XX    ",0,56)
          oled.show()
          hall_1_old = 1
          hall_2_old = 0

    if not hall_1.value() and hall_2.value():
          #if not (hall_1_old == 0 and hall_2_old == 1):
          oled.fill(0)
          oled.text("   XX      XX  ",0,0)
          oled.text("  X  X    XXX  ",0,8)
          oled.text(" X    X  X  X  ",0,16)
          oled.text(" X    X     X  ",0,25)
          oled.text(" X    X     X  ",0,32)
          oled.text(" X    X     X  ",0,40)
          oled.text("  X  X      X  ",0,48)
          oled.text("   XX     XXXXX",0,56)
          oled.show()
          hall_1_old = 0
          hall_2_old = 1


# Definition of first "State Machine"
# that checks the pins of the rotary sensor disc

@asm_pio(autopush=True, push_thresh=32)
def encoder():
    label("start")
    set(x, 0)

    label("loop_high")
    wait(1, pin, 0)         # Wait until in_base pin goes HIGH
    jmp(pin, "CW_1")        # if jmp_pin is low CW rotation
    jmp(x_dec, "CCW_1")     # if not, CCW rotation, Decrement X
    label("CCW_1")
    jmp("loop_low")         # jump to "wait for loop going low"
    label("CW_1")           # CW rotation
    mov(y, invert(x))       # Increment X
    jmp(y_dec, "CW_1a")
    label("CW_1a")
    mov(x, invert(y))


    label("loop_low")
    wait(0, pin, 0)         # Wait until in_base pin goes LOW
    jmp(pin, "CCW_2")       # if jmp_pin is low CCW rotation
    mov(y, invert(x))       # if not, CW rotation Increment X
    jmp(y_dec, "CW_2")
    label("CW_2")
    mov(x, invert(y))
    jmp("loop_high")
    label("CCW_2")          # CCW rotation detected
    jmp(x_dec, "loop_high") # Decrement X
    jmp("loop_high")

    wrap()

sm1 = StateMachine(1, encoder, freq=125_000_000, in_base=Pin(HALL_1_PIN), jmp_pin=Pin(HALL_2_PIN))
sm1.active(1) # Start the mini program

# Definition of second "State Machine"
# that checks for incoming servo control PWM signal
# Note that in_base pin must be identical to jmp_pin !!!

@asm_pio(autopush=True, push_thresh=32)
def servo_PWM_read():
    label("start")
    set(x, 0)

    label("loop")
    set(y, 0)
    mov(y, invert(y))
    wait(1, pin, 0)         # Wait until in_base pin goes HIGH
    label("still_high")
    jmp(y_dec, "plus_1")    # Pin went from LOW to HIGH, start countdown
    label("plus_1")
    jmp(pin, "still_high")  # if jmp_pin still high, continue counting
    mov(y, invert(y))       # Pin went to LOW, no more counting
    mov(x, y)               # Store number of counted loops in X
    jmp("loop")

    wrap()

sm2 = StateMachine(2, servo_PWM_read, freq=125_000_000, in_base=Pin(PWM_READ), jmp_pin=Pin(PWM_READ))
sm2.active(1) # Start the mini program


oled.fill(0) # clear the OLED
oled.text("Pico servo",oled_x,oled_y) # add text at (oled_x,oled_y)
oled.show() # show the new text

# Wait time to start camera recording (set to 0 or uncomment)
# time.sleep(5)

# Enable H bridge
motor_EN.high()

# Get initial states of hall sensors
hall_state = ((hall_state & 3) << 2) | (hall_1.value() << 1) | hall_2.value()


if not CONSTANT_ROTATION: # Servo mode
    # go to initial position
    oled.fill(0)
    oled.text("Going to",oled_x,oled_y)
    oled.text("end switch...",oled_x,oled_y + 20)
    oled.show()

    motor_DIR.low()
    motor_PWM.duty_u16(PWM_END_SWITCH)

    while hall_3.value(): # Wait for end switch going LOW
        if END_SWITCH_LED:
            if hall_3.value():
                led_int.high()
            else:
                led_int.low()

    motor_DIR.low()
    motor_PWM.duty_u16(0)

    oled.fill(0)
    oled.text("End switch",oled_x,oled_y)
    oled.text("reached!",oled_x,oled_y + 20)
    oled.show()

    sm1.exec("set(x, 0)")
    time.sleep(1) # uncomment to make init faster


while True:

  if END_SWITCH_LED:
      if hall_3.value():
          led_int.high()
      else:
          led_int.low()


  if READ_SERVO_PWM and not TEST_SERVO_POS:
      sm2.exec("in_(x, 32)")
      servo_pulse = sm2.get()
      # 61300 = 1.0ms 30650 = 0.5ms
      servo_ms = float(servo_pulse / 61300.0)

      if CONSTANT_ROTATION:
          speed_set = int(float(70 * (servo_ms - 1.5)))
          if abs(speed_set) < 10:  # if speed_set lower 10, stop motor
              speed_set = 0
      else: # Servo mode
          set_point = int(float((servo_ms-0.5) * SET_POINT_180 / 2.0))

  if CONSTANT_ROTATION:

      if pwm_set == 0 and speed_set != 0:
          pwm_set = PWM_MIN

      old_speed_diff = current_speed_diff

      # Detect number of pulses in 0.5s (=500000us)
      sm1.exec("in_(x, 32)")
      encoderStepsOld = 2 * to_int32(sm1.get())

      start_us = time.ticks_us()
      while time.ticks_us() < start_us + 500000:
          i=0

      sm1.exec("in_(x, 32)")
      encoderSteps = 2 * to_int32(sm1.get())

      current_speed = encoderSteps - encoderStepsOld
      current_speed_diff = abs(current_speed - speed_set)

      if DISPLAY_SENSOR_STATE:
          sensors_01()
      else:
          # Print parameters on OLED
          oled.fill(0)
          oled.text("Steps   =" + str(encoderSteps),oled_x,oled_y)
          oled.text("Previous=" + str(encoderStepsOld),oled_x,oled_y + oled_line)
          oled.text("Diff    =" + str(encoderSteps-encoderStepsOld),oled_x,oled_y + 2 * oled_line)
          oled.text("Speed   =" + str(speed_set),oled_x,oled_y + 3 * oled_line)
          oled.text("PWM     =" + str(pwm_set),oled_x,oled_y + 4 * oled_line)
          oled.text("Control =" + str(servo_pulse),oled_x,oled_y + 5 * oled_line)
          oled.show()

          if abs(speed_set) > 0:

              # Check if speed gets closer to set speed or not
              if old_speed_diff > current_speed_diff and abs(current_speed) > 0:
                  OutOfRange = 0
              else:  # Current speed doesn't get closer to set speed
                  OutOfRange += 1

              # Set motor power
              if speed_set > 0: # Clockwise rotation, negative speed_set
                  motor_DIR.high()
                  motor_PWM.duty_u16(MAX_DUTY_16-pwm_set)
              else:
                  motor_DIR.low() # Counterclockwise rotation, positive speed_set
                  motor_PWM.duty_u16(pwm_set)
              if abs(current_speed) < abs(speed_set):# Motor is behind speed
                  pwm_set += 400
              elif abs(current_speed) > abs(speed_set): # Motor is beyond speed
                  pwm_set -= 400
              else: # Motor is within speed range
                  OutOfRange = 0;

          else:             # Stop motor if speed is set to zero
                  motor_DIR.low()
                  motor_PWM.duty_u16(0)
                  OutOfRange = 0;
                  pwm_set = PWM_MIN


      if OutOfRange > 10: # Emergency stop
          pwm_set = 0
          motor_DIR.low()
          motor_PWM.duty_u16(0) # Turn motor OFF

          # Print Warning on OLED
          oled.fill(0)
          oled.text("       XX      ",0,0)
          oled.text("      XXXX     ",0,8)
          oled.text("      XXXX     ",0,16)
          oled.text("      XXXX     ",0,25)
          oled.text("       XX      ",0,32)
          oled.text("       XX      ",0,40)
          oled.text("               ",0,48)
          oled.text("       XX      ",0,56)
          oled.show()

          while True: # Loop forever !!!!!
              i = 0


  else: # Servo mode

      # Get current servo position from "State Machine"
      sm1.exec("in_(x, 32)") # Read current position of sensor disc
      encoderSteps = 2 * to_int32(sm1.get()) # convert to signed integer

      # get current sensor states and fine tune position
      hall_state = (hall_1.value() << 1) | hall_2.value()
      if hall_state == 1:
          encoderSteps += 1
      if hall_state == 2:
          encoderSteps += 1

      # Calculate pwm_set(=motor power):
      # The closer to the set point, the lower the power delivered to the motor
      pwm_set = int(float(PWM_MAX * (abs(set_point - encoderSteps) / PID_PROPORTIONAL)))

      if pwm_set > PWM_MAX:
          pwm_set = PWM_MAX

      if pwm_set < PWM_MIN:
          pwm_set = PWM_MIN

      # What to display on OLED
      if DISPLAY_SENSOR_STATE:
          sensors_01()
      else:
          # Print parameters on OLED
          oled.fill(0)
          oled.text("X=" + str(encoderSteps),oled_x,oled_y)
          oled.text("set point=" + str(set_point),oled_x,oled_y + oled_line)
          oled.text("PWM=" + str(pwm_set),oled_x,oled_y + oled_line * 2)
          oled.text("Control =" + str("{0:.2f}".format(servo_ms))+"ms",oled_x,oled_y + 48)
          oled.show()
          encoderStepsOld = encoderSteps

      # If servo lever is behind set point
      if encoderSteps + SET_POINT_TOLERANCE < set_point:
          motor_DIR.high()
          motor_PWM.duty_u16(MAX_DUTY_16-pwm_set)

      # Else if servo lever is beyond set point
      elif encoderSteps > set_point + SET_POINT_TOLERANCE:
          motor_DIR.low()
          motor_PWM.duty_u16(pwm_set)

      # Servo lever has hit set point
      else:
          motor_DIR.low()
          motor_PWM.duty_u16(0)

          # calculate next servo lever point if TEST_SERVO_POS = True
          if TEST_SERVO_POS:
              new_servo_pos += 1
              if new_servo_pos == 40:
                  new_servo_pos = 0
                  set_point += 100
                  if set_point > SET_POINT_MAX:
                      set_point = 0


